/* **********************************************************
 * Copyright 2006 VMware, Inc.  All rights reserved. -- VMware Confidential
 * **********************************************************/

/*
 * bdhandler.c --
 *
 * Background thread for handling backdoor requests and replies.
 */

/* Must come before any kernel header file. */
#include "driver-config.h"

#include <asm/atomic.h>
#include <asm/semaphore.h>
#include <linux/errno.h>
#include <linux/timer.h>
#include "compat_completion.h"
#include "compat_kernel.h"
#include "compat_list.h"
#include "compat_sched.h"
#include "compat_slab.h"
#include "compat_spinlock.h"
#include "compat_version.h"

/* Must be included after sched.h. */
#include <linux/smp_lock.h>

#include "hgfsBd.h"
#include "hgfsDevLinux.h"
#include "hgfsProto.h"
#include "bdhandler.h"
#include "module.h"
#include "request.h"
#include "vm_assert.h"
#include "vmrpc.h"
#include "eventManager.h"
#include "vm_app.h"

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 4, 9)
int errno;  /* compat_exit() needs global errno variable. */
#endif

static inline void HgfsWakeWaitingClient(HgfsReq *req);
static inline void HgfsCompleteReq(HgfsReq *req,
                                   char const *reply,
                                   uint32 replySize);
static void HgfsSendUnsentReqs(void);
static void HgfsBdPoll(unsigned long data);
static Bool HgfsBdOpenTcloChannel(void);
static void HgfsBdCloseTcloChannel(void);
static HgfsStatus HgfsTcloOplockChange(char const *packetIn,
                                       char *packetOut,
                                       uint32 *packetSize);
static Bool HgfsTcloReset(char const **result,
                          size_t *resultLen,
                          const char *name,
                          const char *args,
                          unsigned int argsSize,
                          void *clientData);
static void HgfsTcloError(void *clientData,
                          char const *status);
static Bool HgfsTcloDispatchReply(char const **result,
                                  size_t *resultLen,
                                  const char *name,
                                  const char *args,
                                  unsigned int argsSize,
                                  void *clientData);
static Bool HgfsTcloDispatchRequest(char const **result,
                                    size_t *resultLen,
                                    const char *name,
                                    const char *args,
                                    unsigned int argsSize,
                                    void *clientData);

/*
 * Private function implementations.
 */

/*
 *----------------------------------------------------------------------
 *
 * HgfsWakeWaitingClient --
 *
 *    Wakes up the client process waiting for the reply to this
 *    request.
 *
 * Results:
 *    None
 *
 * Side effects:
 *    None
 *
 *----------------------------------------------------------------------
 */

static inline void
HgfsWakeWaitingClient(HgfsReq *req)  // IN: Request
{
   ASSERT(req);
   wake_up(&req->queue);
}


/*
 *----------------------------------------------------------------------
 *
 * HgfsCompleteReq --
 *
 *    Copies the reply packet into the request structure and wakes up
 *    the associated client.
 *
 * Results:
 *    None
 *
 * Side effects:
 *    None
 *
 *----------------------------------------------------------------------
 */

static inline void 
HgfsCompleteReq(HgfsReq *req,       // IN: Request
                char const *reply,  // IN: Reply packet
                uint32 replySize)   // IN: Size of reply packet
{
   ASSERT(replySize <= HGFS_PACKET_MAX);

   memcpy(HGFS_REQ_PAYLOAD(req), reply, replySize);
   req->payloadSize = replySize;
   req->state = HGFS_REQ_STATE_COMPLETED;
   list_del_init(&req->list);
   HgfsWakeWaitingClient(req);         
}


/*
 *-----------------------------------------------------------------------------
 *
 * HgfsSendUnsentReqs --
 *
 *      Process the unsent list and send requests to the backdoor.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

static void
HgfsSendUnsentReqs(void)
{
   char const *replyPacket;
   struct list_head *cur, *tmp;
   HgfsReq *req;
   uint32 payloadSize;
   Bool isPacketAsync;

   spin_lock(&hgfsBigLock);
   list_for_each_safe(cur, tmp, &hgfsReqsUnsent) {
      req = list_entry(cur, HgfsReq, list);

      /* 
       * A big "wtf" from the driver is in order. Perhaps by "wtf" I really
       * mean BUG_ON().
       */
      ASSERT(req->state == HGFS_REQ_STATE_UNSENT);
      if (req->state != HGFS_REQ_STATE_UNSENT) {
         LOG(2, (KERN_DEBUG "VMware hgfs: HgfsSendUnsentReqs: Found request "
                 "on unsent list in the wrong state, ignoring\n"));
         continue;
      }

      ASSERT(req->payloadSize <= HGFS_PACKET_MAX);
      isPacketAsync = hgfsUseAsyncBackdoor;
     retry:
      payloadSize = req->payloadSize;
      LOG(8, (KERN_DEBUG "VMware hgfs: HgfsSendUnsentReqs: Sending packet "
              "over backdoor\n"));

      /* Write the appropriate command into the packet. */
      if (isPacketAsync) {
         memcpy(req->packet, HGFS_ASYNC_REQUEST_CLIENT_CMD, 
                HGFS_ASYNC_REQUEST_CLIENT_CMD_LEN);
      } else {
         memcpy(req->packet, HGFS_SYNC_REQREP_CLIENT_CMD, 
                HGFS_SYNC_REQREP_CLIENT_CMD_LEN);         
      }

      /* 
       * We should attempt to reopen the backdoor channel with every request, 
       * because the HGFS server in the host can be enabled or disabled at any 
       * time.
       */
      if (!HgfsBd_OpenBackdoor(&hgfsRpcOut)) {
         req->state = HGFS_REQ_STATE_ERROR;
         list_del_init(&req->list);
         printk(KERN_WARNING "VMware hgfs: HGFS is disabled in the host\n");
         HgfsWakeWaitingClient(req);
      } else if (HgfsBd_Dispatch(hgfsRpcOut, HGFS_REQ_PAYLOAD(req),
                                 &payloadSize, &replyPacket) == 0) {

         /* 
          * Request sent successfully. If it was asynchronous, just put us on 
          * the sent list. Otherwise, copy the reply and wake up the client.
          */
         if (isPacketAsync) {
            req->state = HGFS_REQ_STATE_SENT;
            list_del(&req->list);
            list_add_tail(&req->list, &hgfsReqsSent);
         } else {
            HgfsCompleteReq(req, replyPacket, payloadSize);
         }
         LOG(8, (KERN_DEBUG "VMware hgfs: HgfsSendUnsentReqs: Backdoor "
                 "reply received\n"));
      } else if (isPacketAsync) {
         /*
          * We failed on an asynchronous request. Fallback to synchronous
          * and retry the send.
          *
          * The only reason we're using two variables to track this is to
          * prevent races on hgfsUseAsyncBackdoor.
          */
         hgfsUseAsyncBackdoor = isPacketAsync = FALSE;
         LOG(8, (KERN_DEBUG "VMware hgfs: HgfsSendUnsentReqs: Server does not "
                 "support asynchronous HGFS, retrying synchronously\n"));
         goto retry;
      } else {
         
         /* Pass the error into the request. */
         req->state = HGFS_REQ_STATE_ERROR;
         list_del_init(&req->list);
         LOG(8, (KERN_DEBUG "VMware hgfs: HgfsSendUnsentReqs: Backdoor "
                 "error\n"));
         HgfsWakeWaitingClient(req);

         /*
          * If the channel was previously open, make sure it's dead and gone
          * now. We do this because subsequent requests deserve a chance to
          * reopen it.
          */
         HgfsBd_CloseBackdoor(&hgfsRpcOut);
      }
   }
   spin_unlock(&hgfsBigLock);
}


/*
 *----------------------------------------------------------------------
 *
 * HgfsBdPoll --
 *
 *    This is our poll loop timer callback. Just wake up the bdhandler
 *    thread.
 *
 * Results:
 *    None.
 *
 * Side effects:
 *    None.
 *
 *----------------------------------------------------------------------
 */

static void
HgfsBdPoll(unsigned long data)  // Unused
{
   set_bit(HGFS_REQ_THREAD_POLL, &hgfsReqThreadFlags);
   wake_up_interruptible(&hgfsReqThreadWait);
}


/*
 *-----------------------------------------------------------------------------
 *
 * HgfsBdOpenTcloChannel --
 *
 *      Attempt to construct and start an RpcIn TCLO channel to the host.
 *
 * Results:
 *      TRUE if channel was successfully started.
 *      FALSE otherwise.
 *
 * Side effects:
 *      May start a TCLO channel to the host.
 *
 *-----------------------------------------------------------------------------
 */

static Bool
HgfsBdOpenTcloChannel(void)
{
   Bool success = FALSE;

   hgfsEventQueue = EventManager_Init();
   if (hgfsEventQueue == NULL) {
      LOG(6, (KERN_WARNING "VMware hgfs: failed to create event queue\n"));
      goto exit;
   }
   hgfsRpcIn = RpcIn_Construct(hgfsEventQueue);
   if (hgfsRpcIn == NULL) {
      LOG(6, (KERN_WARNING "VMware hgfs: failed to create TCLO handler\n"));
      goto exit;
   }
   if (!RpcIn_start(hgfsRpcIn, RPCIN_POLL_TIME, 
                    HgfsTcloReset, NULL, 
                    HgfsTcloError, NULL)) {
      LOG(6, (KERN_WARNING "VMware hgfs: failed to start TCLO handler\n"));
      goto exit;
   }
   
   RpcIn_RegisterCallback(hgfsRpcIn, HGFS_ASYNC_REPLY_CMD, 
                          HgfsTcloDispatchReply, NULL);
   RpcIn_RegisterCallback(hgfsRpcIn, HGFS_SYNC_REQREP_CMD, 
                          HgfsTcloDispatchRequest, NULL);
   
   success = TRUE;
  exit:
   if (success == FALSE) {
      if (hgfsRpcIn != NULL) {
         RpcIn_stop(hgfsRpcIn);
         RpcIn_Destruct(hgfsRpcIn);
      }
      if (hgfsEventQueue != NULL) {
         EventManager_Destroy(hgfsEventQueue);
      }
   }
   return success;
}


/*
 *-----------------------------------------------------------------------------
 *
 * HgfsBdCloseTcloChannel --
 *
 *      Attempts to stop and free a TCLO channel to the host.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      Possibly stops a TCLO channel.
 *
 *-----------------------------------------------------------------------------
 */
static void
HgfsBdCloseTcloChannel(void)
{
   /* Stop the TCLO loop. */
   if (!RpcIn_stop(hgfsRpcIn)) {
      printk(KERN_WARNING "VMware hgfs: failed to stop the TCLO loop\n");
   }
   RpcIn_Destruct(hgfsRpcIn);
   EventManager_Destroy(hgfsEventQueue);
}


/*
 *-----------------------------------------------------------------------------
 *
 * HgfsTcloOplockChange --
 *
 *      Callback function that handles a change in the oplock status of a file
 *      as prescribed by the HGFS server. This typically occurs when another 
 *      client is trying to open a file that is protected by the oplock.
 *
 *      As oplocks do not yet do anything significant in the driver, simply
 *      return success.
 *
 * Results:
 *      TRUE always.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

static HgfsStatus
HgfsTcloOplockChange(char const *packetIn, // IN:     Incoming request packet
                     char *packetOut,      // OUT:    Outgoing reply packet
                     uint32 *packetSize)   // IN/OUT: Size of packet
{
   HgfsRequestServerLockChange *request;
   HgfsReplyServerLockChange *reply;
   HgfsHandle serverFileId;
   HgfsServerLock newServerLock;

   request = (HgfsRequestServerLockChange *)packetIn;
   reply = (HgfsReplyServerLockChange *)packetOut;

   serverFileId = request->file;
   newServerLock = request->newServerLock;
   LOG(4, (KERN_DEBUG "VMware hgfs: HgfsTcloOplockChange: Received oplock "
           "change request to lock %d for file with handle %d\n", 
           newServerLock, serverFileId));

   /* Tell the server that we've broken our lock instead of degrading. */
   reply->serverLock = HGFS_LOCK_NONE;
   *packetSize = sizeof *reply;
   return HGFS_STATUS_SUCCESS;
}


/*
 *-----------------------------------------------------------------------------
 *
 * HgfsTcloReset --
 *
 *      We'll receive this TCLO message on a power state change. Some
 *      things we'd like to do:
 *      -Reset ops with multiple opcodes to V2.
 *      -Reset the use of the asynchronous backdoor.
 *      -Reset the HGFS backdoor.
 *
 * Results:
 *      Always TRUE.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

static Bool 
HgfsTcloReset(char const **result,   // OUT
              size_t *resultLen,     // OUT
              const char *name,      // IN: Ignored      
              const char *args,      // IN: Ignored
              unsigned int argsSize, // IN: Ignored
              void *clientData)      // Unused
{
   LOG(2, (KERN_DEBUG "VMware hgfs: HgfsTcloReset: Received 'reset'\n"));
   HgfsResetOps();
   HgfsBd_CloseBackdoor(&hgfsRpcOut);
   HgfsBd_OpenBackdoor(&hgfsRpcOut);
   return RpcIn_SetRetVals(result, resultLen, "ATR " TOOLS_HGFS_NAME, TRUE);
}


/*
 *-----------------------------------------------------------------------------
 *
 * HgfsTcloError --
 *
 *      An error occurred in the RpcIn loop.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

static void 
HgfsTcloError(void *clientData,    // Ignored
              char const *status)  // Ignored
{
   /* XXX: Mark some global state so we'll restart the RpcIn loop. */
}


/*
 *-----------------------------------------------------------------------------
 *
 * HgfsTcloDispatchReply --
 *
 *      We got a reply to an HGFS packet. Find the matching request for that
 *      reply in the sent list, stuff the reply into the request's packet
 *      field, and wake up its client.
 *
 * Results:
 *      TRUE if reply was sane, even if impertinent.
 *      FALSE if the reply was malformed.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

static Bool 
HgfsTcloDispatchReply(char const **result,   // OUT
                      size_t *resultLen,     // OUT
                      const char *name,      // IN: Ignored
                      const char *args,      // IN: Reply packet
                      unsigned int argsSize, // IN: Size of reply packet
                      void *clientData)      // Unused
{
   struct list_head *cur, *tmp;
   const char *packet;
   uint32 packetSize;
   HgfsReq *req;
   HgfsRequest *reqHeader;
   HgfsReply *repHeader;

   /* Some sanity checks. */
   if (argsSize == 0) {
      return RpcIn_SetRetVals(result, resultLen, "1 argument required", FALSE);
   }
   packet = args + 1;
   packetSize = argsSize - 1;
   repHeader = (HgfsReply *)packet;
   if (args[0] != ' ' ||
       packetSize < sizeof *repHeader ||
       packetSize > HGFS_PACKET_MAX) {
      LOG(2, (KERN_DEBUG "VMware hgfs: HgfsTcloDispatchReply: Received "
              "malformed reply, bailing\n"));
      return RpcIn_SetRetVals(result, resultLen, "Malformed packet", FALSE);
   }

   LOG(8, (KERN_DEBUG "VMware hgfs: HgfsTcloDispatchReply: Got reply with id "
           "%d\n", repHeader->id));

   /* Look at all of our sent requests. */
   spin_lock(&hgfsBigLock);
   list_for_each_safe(cur, tmp, &hgfsReqsSent) {
      req = list_entry(cur, HgfsReq, list);

      LOG(8, (KERN_DEBUG "VMware hgfs: HgfsTcloDispatchReply: Looking at "
              "request id %d (%s)\n", req->id, req->id == repHeader->id ? 
              "match" : "no match"));

      /* See "wtf" in HgfsSendUnsentReqs. */
      ASSERT(req->state == HGFS_REQ_STATE_SENT);
      if (req->state != HGFS_REQ_STATE_SENT) {
         LOG(2, (KERN_DEBUG "VMware hgfs: HgfsTcloDispatchReply: Found "
                 "request on sent list in the wrong state, ignoring\n"));
         continue;
      }

      reqHeader = (HgfsRequest *)(HGFS_REQ_PAYLOAD(req));
      if (reqHeader->id == repHeader->id) {
         HgfsCompleteReq(req, packet, packetSize);
         break;
      }
   }
   spin_unlock(&hgfsBigLock);

   /* 
    * It's possible that the reply had no matching request. No big deal; we
    * probably abandoned it earlier.
    */
   return RpcIn_SetRetVals(result, resultLen, "", TRUE);
}


/*
 *-----------------------------------------------------------------------------
 *
 * HgfsTcloDispatchRequest --
 *
 *      The server sent us an out-of-band request. Dispatch it to the
 *      appropriate handler function.
 *
 * Results:
 *      TRUE if request was sane.
 *      FALSE if request was malformed.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

static Bool 
HgfsTcloDispatchRequest(char const **result,   // OUT
                        size_t *resultLen,     // OUT
                        const char *name,      // IN: Ignored
                        const char *args,      // IN: Reply packet
                        unsigned int argsSize, // IN: Size of reply packet
                        void *clientData)      // Unused
{
   static char packetOut[HGFS_PACKET_MAX];
   const char *packetIn;
   uint32 packetSize;
   HgfsRequest *reqHeader;
   HgfsReply *repHeader;
   HgfsHandle id;
   HgfsOp op;
   HgfsStatus status;

   /* Some sanity checks. */
   if (argsSize == 0) {
      return RpcIn_SetRetVals(result, resultLen, "1 argument required", FALSE);
   }
   packetIn = args + 1;
   packetSize = argsSize - 1;
   reqHeader = (HgfsRequest *)packetIn;
   repHeader = (HgfsReply *)packetOut;
   if (args[0] != ' ' ||
       packetSize < sizeof *reqHeader ||
       packetSize > HGFS_PACKET_MAX) {
      LOG(2, (KERN_DEBUG "VMware hgfs: HgfsTcloDispatchRequest: Received "
              "malformed request, bailing\n"));
      return RpcIn_SetRetVals(result, resultLen, "Malformed packet", FALSE);
   }

   LOG(8, (KERN_DEBUG "VMware hgfs: HgfsTcloDispatchRequest: Got request with "
           "id %d\n", reqHeader->id));
   id = reqHeader->id;
   op = reqHeader->op;

   switch (op) {
   case HGFS_OP_SERVER_LOCK_CHANGE:
      status = HgfsTcloOplockChange(packetIn, packetOut, &packetSize);
      break;
   default:
      status = HGFS_STATUS_PROTOCOL_ERROR;
      break;
   }

   /* Finish setting up the reply and pass it down to the TCLO channel. */
   repHeader->id = id;
   repHeader->status = status;
   *result = packetOut;

   /* 
    * Handler functions need not set packetSize on error, so we'll do it
    * for them.
    */
   if (status == HGFS_STATUS_SUCCESS) {
      *resultLen = packetSize;
   } else {
      *resultLen = sizeof *repHeader;
   }
   return TRUE;
}


/*
 * Public function implementations.
 */

/*
 *-----------------------------------------------------------------------------
 *
 * HgfsResetOps --
 *
 *      Reset ops with more than one opcode back to the desired opcode. Also
 *      allow the driver to attempt asynchronous traffic.
 *
 * Results:
 *      None.
 *
 * Side effects:
 *      None.
 *
 *-----------------------------------------------------------------------------
 */

void
HgfsResetOps(void)
{
   atomic_set(&hgfsVersionOpen, HGFS_OP_OPEN_V2);
   atomic_set(&hgfsVersionGetattr, HGFS_OP_GETATTR_V2);
   atomic_set(&hgfsVersionSetattr, HGFS_OP_SETATTR_V2);
   atomic_set(&hgfsVersionSearchRead, HGFS_OP_SEARCH_READ_V2);
   atomic_set(&hgfsVersionCreateDir, HGFS_OP_CREATE_DIR_V2);

   /* XXX: This is currently FALSE for performance reasons. */
   hgfsUseAsyncBackdoor = FALSE;
}


/*
 *----------------------------------------------------------------------
 *
 * HgfsBdHandler --
 *
 *    Function run in background thread to pick up HGFS requests from
 *    the filesystem half of the driver, send them over the backdoor,
 *    get replies, and send them back to the filesystem.
 *
 * Results:
 *    Always returns zero.
 *
 * Side effects:
 *    Processes entries from hgfsReqQ.
 *
 *----------------------------------------------------------------------
 */

int 
HgfsBdHandler(void *data)
{
   Bool gotTcloChannel;
   struct timer_list tcloTimer;
   struct timespec ct;
   uint64 sleepUsecs = 0;

   LOG(6, (KERN_DEBUG "VMware hgfs: HgfsBdHandler: Thread starting\n"));
   compat_daemonize(HGFS_NAME);

   /* 
    * Try and start the TCLO loop. If we fail, we'll run in degraded mode
    * (no power state change notifications, no asynchronous backdoor).
    *
    * XXX: In the future, we ought to retry opening the TCLO channel using
    * a timer.
    */
   LOG(6, (KERN_DEBUG "VMware hgfs: HgfsBdHandler: Starting TCLO loop\n"));
   gotTcloChannel = HgfsBdOpenTcloChannel();
   if (gotTcloChannel) {
      init_timer(&tcloTimer);
      tcloTimer.expires = jiffies + RPCIN_POLL_TICKS;
      tcloTimer.data = (unsigned long)&tcloTimer;
      tcloTimer.function = HgfsBdPoll;
      add_timer(&tcloTimer);
   } else {
      printk(KERN_WARNING "VMware hgfs: TCLO loop failed to start. "
             "Running in degraded mode\n");
   }

   for (;;) {

      /* Sleep, waiting for a request, a poll time, or exit. */
      wait_event_interruptible(hgfsReqThreadWait,
                               test_bit(HGFS_REQ_THREAD_SEND, 
                                        &hgfsReqThreadFlags) ||
                               test_bit(HGFS_REQ_THREAD_POLL, 
                                        &hgfsReqThreadFlags) ||
                               test_bit(HGFS_REQ_THREAD_EXIT, 
                                        &hgfsReqThreadFlags));
      
      /* 
       * First, check for suspend. I'm not convinced that this actually
       * has to come first, but whatever.
       */
      compat_try_to_freeze();

      /* Send outgoing requests. */
      if (test_and_clear_bit(HGFS_REQ_THREAD_SEND, &hgfsReqThreadFlags)) {
         LOG(8, (KERN_DEBUG "VMware hgfs: HgfsBdHandler: Sending requests\n"));
         HgfsSendUnsentReqs();
      }

      /* Poll for a packet. */
      if (test_and_clear_bit(HGFS_REQ_THREAD_POLL, &hgfsReqThreadFlags)) {
         if (EventManager_ProcessNext(hgfsEventQueue, &sleepUsecs) == -1) {

            /* We don't reschedule the timer, so the polling ends here. */
            printk(KERN_WARNING "VMware hgfs: Unexpected failure in event "
                   "queue. Entering degraded mode.\n");
         } else {
            
            /* XXX: Ugh, the lack of __udivdi3 on 2.6.13+ is killing me. */
            ct.tv_sec = 0;
            ct.tv_nsec = sleepUsecs * 1000;
            
            /* Timer expiration is in terms of ticks, not us. */
            mod_timer(&tcloTimer, jiffies + timespec_to_jiffies(&ct));
         }
      }

      /* Kill yourself. */
      if (test_and_clear_bit(HGFS_REQ_THREAD_EXIT, &hgfsReqThreadFlags)) { 
         LOG(6, (KERN_DEBUG "VMware hgfs: HgfsBdHandler: Told to exit\n"));
         break;
      }
   }

   LOG(6, (KERN_DEBUG "VMware hgfs: HgfsBdHandler: Closing backdoor\n"));
   HgfsBd_CloseBackdoor(&hgfsRpcOut);
   if (gotTcloChannel) {
      LOG(6, (KERN_DEBUG "VMware hgfs: HgfsBdHandler: Stopping TCLO loop\n"));
      del_timer_sync(&tcloTimer);
      HgfsBdCloseTcloChannel();
   }

   LOG(6, (KERN_DEBUG "VMware hgfs: HgfsBdHandler: Thread exiting\n"));

   /* Signal our parent that we're exiting, and exit, all at once. */
   compat_complete_and_exit(&hgfsReqThreadDone, 0);
   NOT_REACHED();
}
